import _io
import json
import platform
import random
from dataclasses import dataclass
from sys import argv
import sys
import time
from color_class import Color
import os

SIZE_TYPE = (int, int)
WINDOWS = "Windows"
system = platform.system()

GAME_OVER = """
  ▄▄█▀▀▀█▄█                                       ▄▄█▀▀██▄                             
▄██▀     ▀█                                     ▄██▀    ▀██▄                           
██▀       ▀ ▄█▀██▄ ▀████████▄█████▄   ▄▄█▀██    ██▀      ▀████▀   ▀██▀  ▄▄█▀██▀███▄███ 
██         ██   ██   ██    ██    ██  ▄█▀   ██   ██        ██ ██   ▄█   ▄█▀   ██ ██▀ ▀▀ 
██▄    ▀████▄█████   ██    ██    ██  ██▀▀▀▀▀▀   ██▄      ▄██  ██ ▄█    ██▀▀▀▀▀▀ ██     
▀██▄     ████   ██   ██    ██    ██  ██▄    ▄   ▀██▄    ▄██▀   ███     ██▄    ▄ ██     
  ▀▀███████▀████▀██▄████  ████  ████▄ ▀█████▀     ▀▀████▀▀      █       ▀█████▀████▄   
"""

YOU_WIN = """
▀███▀   ▀██▀                       ▀████▀     █     ▀███▀████▀███▄   ▀███▀█ 
  ███   ▄█                           ▀██     ▄██     ▄█   ██   ███▄    █ ██ 
   ███ ▄█    ▄██▀██▄▀███  ▀███        ██▄   ▄███▄   ▄█    ██   █ ███   █ ██ 
    ████    ██▀   ▀██ ██    ██         ██▄  █▀ ██▄  █▀    ██   █  ▀██▄ █ ██ 
     ██     ██     ██ ██    ██         ▀██ █▀  ▀██ █▀     ██   █   ▀██▄█ ▀▀ 
     ██     ██▄   ▄██ ██    ██          ▄██▄    ▄██▄      ██   █     ███ ▄▄ 
   ▄████▄    ▀█████▀  ▀████▀███▄         ██      ██     ▄████▄███▄    ██ ██ 
"""


def clear():
    os.system("cls" if system == WINDOWS else r'printf "\033c"')


def resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath(".")

    return os.path.join(base_path, relative_path)


def args(*flags: str):
    arg = argv[1:]
    if any(i in arg for i in flags):
        return True
    for f in flags:
        if len(f) == 2 and f[0] == '-':
            return f[1] in ''.join(filter(lambda x: len(x) >= 2 and x[0] == '-' and x[1:].isalpha(), arg))
    return False


class Settings:
    ANIMATION = "animation"
    DIFFICULT = "difficult"
    SIMPLE_SYMBOLS = "simple_symbols"
    COLOR = "color"
    CHECK_SIZE = "check_size"

    SETTINGS_PATH = resource_path("settings.json")  # type: str
    DATA = json.load(open(SETTINGS_PATH))  # type: dict

    SYMBOLS = {
        "hero": Color("@", Color.GREEN),
        "monster": Color("M", Color.RED),
        "wall": Color("W", Color.WHITE, Color.ON_WHITE),
        "medicare": Color("+", Color.PURPLE),
        "portal": Color("0", Color.BLUE),
    }

    @classmethod
    def get_param(cls, key: str, *flags: str):
        return args(*flags) or cls.DATA.get(key, False)

    @classmethod
    def get_animation(cls):
        return cls.get_param(cls.ANIMATION, "-f", "--fast")

    @classmethod
    def get_color(cls):
        return cls.get_param(cls.COLOR, "-c", "--no-color")

    @classmethod
    def save_settings(cls):
        json.dump(cls.DATA, open(cls.SETTINGS_PATH, 'w'), indent=2, ensure_ascii=True)

    @classmethod
    def set_any_bool(cls, key: str, value: bool):
        cls.DATA[key] = bool(value)
        cls.save_settings()

    @classmethod
    def set_animation(cls, value: bool):
        cls.set_any_bool(cls.ANIMATION, value)

    @classmethod
    def set_color(cls, value: bool):
        cls.set_any_bool(cls.COLOR, value)
        Color.DE_COLOR = not cls.get_color()


Color.DE_COLOR = not Settings.get_color()

DOC = """Добро пожаловать в игру "%s")
Правила игры очень просты:
    • Вам просто нужно сражаться с мострами и пройти как можно глубже в "подземелье")
    • Получить удовольствие от игры (если это вообще реально)

Управление: 
    • Перемещение происходит по средсвам кнопок WASD и Enter для ввода
    • Во время битвы Вам нужно атаковать (A) и блокироваться (B)

Начнём играть?) (Д/н): """


def set_console_size(cols: int = None, lines: int = None):
    l, c = get_size()
    os.system("mode con: cols=%s lines=%s" % (cols or c, lines or l))


def deep_copy(args):
    return [deep_copy(i) if type(i) == list else i for i in args]


class Entity:
    symbol = Color(" ")
    empty = True

    def __init_subclass__(cls, **kwargs):
        cls.empty = False

    def __str__(self):
        return str(self.symbol)


class Hero(Entity):
    symbol = Settings.SYMBOLS["hero"]

    def __init__(self, hp=20):
        self.hp = hp
        self.max_hp = 30
        self.coords = (0, 0)
        self.attack = 5


class Monster(Entity):
    symbol = Settings.SYMBOLS["monster"]

    def __init__(self, hp=20):
        self.hp = hp
        self.coords = (0, 0)
        self.attack = 3


class Wall(Entity):
    symbol = Settings.SYMBOLS["wall"]


class Medicare(Entity):
    symbol = Settings.SYMBOLS["medicare"]
    hp = 3


class Portal(Entity):
    symbol = Settings.SYMBOLS["portal"]


class Scene:
    DIRECTIONS = {
        "w": (-1, 0),
        "a": (0, -1),
        "s": (1, 0),
        "d": (0, 1),
    }

    def __init__(self, map):
        if type(map) == _io.TextIOWrapper:
            with map as f:
                map = [[j for j in i if j != "\n"] for i in f.readlines()]
        self.map = deep_copy(map)
        self.monsters = 0
        self.level = 0
        self.hero = None
        for y in range(len(self.map)):
            for x in range(len(self.map[y])):
                cell = self.map[y][x]
                if type(cell) == Hero or cell == Color.de_color(Hero.symbol):
                    if cell == Color.de_color(Hero.symbol):
                        self.map[y][x] = Hero()
                    self.hero = self.map[y][x]
                    self.hero.coords = (y, x)
                elif type(cell) == Monster or cell == Color.de_color(Monster.symbol):
                    if cell == Color.de_color(Monster.symbol):
                        self.map[y][x] = Monster()
                    self.map[y][x].coords = (y, x)
                    self.monsters += 1
                elif type(cell) == Wall or cell == Color.de_color(Wall.symbol):
                    self.map[y][x] = Wall()
                elif type(cell) == Portal or cell == Color.de_color(Portal.symbol):
                    self.map[y][x] = Portal()
                elif type(cell) == Medicare or cell == Color.de_color(Medicare.symbol):
                    self.map[y][x] = Medicare()
                else:
                    self.map[y][x] = Entity()

    def move(self, direction: str):
        c = self.hero.coords
        d = self.DIRECTIONS[direction]
        m = c[0] + d[0], c[1] + d[1]
        target = self.map[m[0]][m[1]]
        if target.empty or type(target) == Medicare:
            if type(target) == Medicare:
                if self.hero.hp + target.hp > self.hero.max_hp:
                    self.hero.hp = self.hero.max_hp
                else:
                    self.hero.hp += target.hp
            self.map[c[0]][c[1]] = Entity()
            self.map[m[0]][m[1]] = self.hero
            self.hero.coords = m
        elif type(target) == Portal and not self.monsters:
            return target
        elif type(target) == Monster:
            return Fight(self.hero, target, self.level).render()
        return self.hero

    def render(self):
        health_text = str(Color("Здоровье: %s" % self.hero.hp, Color.GREEN))
        monsters_text = str(Color("Мостры: %s" % self.monsters, Color.RED))

        map_rendered = ""
        cols, lines = 0, 0
        for y in self.map:
            for x in y:
                if type(x) in (Monster, Hero, Medicare, Portal):
                    map_rendered += " "
                else:
                    map_rendered += str(x)

                map_rendered += str(Wall() if type(x) == Portal and self.monsters else x)
                cols += 2
            map_rendered += "\n"
            lines += 1
        cols //= lines

        header = health_text + " " * (cols - len(Color.de_color(health_text)) - len(Color.de_color(monsters_text))) + \
                 monsters_text + "\n"
        while True:
            clear()
            set_console_size(cols, lines + 6)
            print(header, center_text((0, cols), "Уровень: %s" % self.level),
                  map_rendered, "", "Введите ход: ", sep="\n", end='')
            direction = input().lower().strip()
            if direction in self.DIRECTIONS:
                move = self.move(direction)
                if type(move) == Monster and move.hp <= 0:
                    self.map[move.coords[0]][move.coords[1]] = Entity()
                    self.monsters -= 1
                return move


def get_size():
    s = os.get_terminal_size()
    return s.lines, s.columns


def center_text(size: SIZE_TYPE, text: str, add_right=True):
    y, x = size
    de_colored_text = Color.de_color(text)
    space_l = " " * (x // 2 - len(de_colored_text) // 2)
    space_r = " " * (x - len(space_l) - len(de_colored_text))
    return space_l + text + (space_r if add_right else '')


def slow_print(string: str, delay: float = 0.02):
    y, x = get_size()
    r = ''
    if not Settings.get_animation() or not delay:
        print(string, end='', flush=True)
        return

    for s in string:
        r += s
        if len(r) == x - 1:
            print('\r' + r, end='', flush=True)
            r = ""
            time.sleep(delay)
            print(flush=True)
        elif s == "\n":
            print(flush=True)
            r = ""
        else:
            print('\r' + r, end='', flush=True)
            if s != ' ':
                time.sleep(delay)


def if_yes(answer: str = None):
    if answer is None:
        answer = input()
    return answer.lower() in ("", "да", "д", "y", "yes")


@dataclass
class Fight:
    hero: Hero
    monster: Monster
    level: int
    status: Color = Color("Ваш ход первый")
    STATUS = {"a": "атаковали", "b": "парировали"}

    def render(self):
        health_text = str(Color("Ваше здоровье: %s" % self.hero.hp, Color.GREEN))
        monsters_text = str(Color("Здоровье монстра: %s" % self.monster.hp, Color.RED))
        lines, cols = get_size()

        header = health_text + " " * (cols - len(Color.de_color(health_text)) - len(Color.de_color(monsters_text))) + \
                 monsters_text

        while True:
            clear()
            set_console_size(cols, 6)
            print(header, center_text((0, cols), "Уровень: %s" % self.level), "", self.status,
                  "", "Введите ход (a, b): ", sep="\n", end='')
            motion = input().lower().strip()
            monster_motion = ("a", "b")[random.randint(0, 1)]
            if motion == monster_motion:
                self.status = Color("Вы и монстр %s) Попробуйте снова" % self.STATUS[motion], Color.YELLOW)
            elif motion == "a":
                self.status = Color("Вы успешно атаковали монстра на %s HP!" % self.hero.attack, Color.GREEN)
                self.monster.hp -= self.hero.attack
            else:
                self.status = Color("Вас атаковали на %s HP!" % self.monster.attack, Color.RED)
                self.hero.hp -= self.monster.attack
            if self.hero.hp <= 0:
                return self.hero
            elif self.monster.hp <= 0:
                return self.monster


